
;;  Paradigmata Programovani 2B, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;; 1. CAST: IMPLICITNI DEFINICE PROUDU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;; dalsi ukazky implicitnich proudu

;; proud faktorialu
(define fak-stream
  (cons-stream 1
    (stream-map *
                fak-stream
                (stream-cdr naturals))))

(display-stream fak-stream 10)

;; proud Fibonacciho cisel
(define fib-stream
  (cons-stream 1
    (cons-stream 1
      (stream-map +
                  fib-stream
                  (stream-cdr fib-stream)))))

(display-stream fib-stream 10)

;;;;; vytvareni nekonecnych proudu: build-stream
(define build-stream
  (lambda (f)
    (let proc ((i 0))
      (cons-stream (f i)
                   (proc (+ i 1))))))

;; priklady
(define ones (build-stream (lambda (i) 1)))
(define naturals (build-stream (lambda (i) (+ i 1))))

;;;;; vytvareni nekonecnych proudu z nekonecnych produdu
;;;;; aplikaci nekolika funkci na kazdy prvek
(define expand-stream
  (lambda (stream . modifiers)
    (let next ((pending modifiers))
      (if (null? pending)
          (apply expand-stream 
		 (stream-cdr stream) modifiers)
          (cons-stream ((car pending) (stream-car stream))
                       (next (cdr pending)))))))

(display-stream (proc ones -) 20)        ; to same jako stream-map
(display-stream (proc naturals
                      (lambda (x) (* x x))
                      -
                      +) 20)                ; 1 -1 1 4 -2 2 9 -3 3 16 -4 4 ...

(expand-stream ones - +)      ; -1 1 -1 1 ...
(expand-stream ones + -)      ; 1 -1 1 -1 ...
(expand-stream naturals + -)  ; 1 -1 2 -2 3 -3 4 -4 5 -5 ...

;; vytvoreni proudu celych cisez
(define integers
  (build-stream (lambda (i)
                  (if (= i 0)
                      0
                      ((if (even? i) + -)
                       (quotient (+ i 1) 2))))))

(display-stream integers 10)

;; nebo pouzitim streamu prirozenych a expand-stream
(define integers
  (cons-stream 0 (expand-stream naturals - +)))

;; problem: muzeme vytvorit streamy racionalnich / realnych cisel?
;; odpoved (mozna prekvapiva)
;; 1) racionalnich ano
;; 2) realnych ne

;; stream paru (1 . 1) (2 . 1) (1 . 2) (3 . 1) (2 . 2) (1 . 3) ...
(define pairs
  (let next ((sum 2)
             (a 1)
             (b 1))
    (cons-stream (cons a b)
                 (if (= a 1)
                     (next (+ sum 1) sum 1)
                     (next sum (- a 1) (+ b 1))))))

;; stream zlomku 1/1 2/1 1/2 3/1 2/2 1/3
;; obsahuje vsechna kladna racionalni cisla, nektera ovsem nasobne
;; reba 1/1 je totez co 2/2 a podobne
(define fractions
  (stream-map (lambda (x)
                (/ (car x) (cdr x)))
              pairs))

;; duplicity odstranime filtraci
(define positive-rationals
  (let next ((f-stream fractions))
    (cons-stream
      (stream-car f-stream)
      (next (stream-filter
             (lambda (x)
                (not (= x (stream-car f-stream))))
             f-stream)))))

;; a konecne: stream vsech racionalnich cisel
(define rationals
  (cons-stream 0 (expand-stream positive-rationals - +)))

(display-stream rationals 20)

;; Poznamka: pouziti nekterych procedur na nekonecne streamy nema smysl
;; napriklad stream-length (delka streamu je z povahy streamu ,,blbost'')

;; stream append2 (ma jedine smysl, pokud je prvni stream konecny)
(define stream-append2
  (lambda (s1 s2)
    (stream-foldr (lambda (x y)
                    (cons-stream x (force y)))
                  s2
                  s1)))

;(stream-append ones (stream-map - ones))
;(stream-length ones)

;;;;; dalsi typicke tvoritko na streamy:
;;;;; iterate: seed f -> stream (f seed) (f (f seed)) (f (f (f seed))) ...
(define iterate
  (lambda (f seed)
    (let ((value (f seed)))
      (cons-stream value
                   (iterate f value)))))

(display-stream (iterate - 1) 20)
(display-stream (iterate (lambda (x) (* 2 x)) 1) 20)

;; stream prvocisel (Erathostenovo sito)
(define primes
  (let next ((n-stream (stream-cdr naturals)))
    (cons-stream
      (stream-car n-stream)
      (next (stream-filter
             (lambda (n)
                (> (modulo n (stream-car n-stream)) 0))
             n-stream)))))

(display-stream primes 100)

;; 2. CAST: RIZENI VSTUPNICH/VYSTUPNICH OPERACI POMOCI PROUDU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Kdy pouzit proudy?
;;
;; 1. kdyz ptrebujeme ridit prubeh vypoctu pomoci dat
;; 2. kdyz velikost dat, ktera chceme zpracovavat neni dopredu znama
;;    a neni ani dopredu znamo, kolik dat budeme potrebovat zpracovat
;;    (priklad: pokousime se najit retezec odpovidajici zadanemu vzoru
;;    ve velkem souboru dat -- nema smysl tahat soubor do pameti naraz cely,
;;    protoze retezec muzeme najit treba nekde na zacatku (nebo taky ne)

;; Typicke pouziti proudu: rizeni vstupne/vystupnich operaci (pouziva mj. C++)

;; Scheme: porty (lze si predstavit jako identifikatory otevrenych souboru)
;; cteni a zapis pomoci portu (port je nepovinny argument)

(define p (open-output-file "soubor.txt"))        ; otevre soubor + vrati port
(display "Ahoj svete!" p)                        ; zapis retezec do souboru
(newline p)                                        ; zapise znak ,,novy radek''
(display (map - '(1 2 3 4)) p)                        ; zapis retezec do souboru
(newline p)                                        ; zapise znak ,,novy radek''
(close-output-port p)                                ; uzavre vystupni port

(define p (open-input-file "soubor.txt"))        ; otevre soubor pro cteni
(display (read p))        ; nacte S-vyraz a zobrazi prislusny element
(display (read p))
(display (read p))
(close-input-port p)        ; konec vstupu

;; pro vstupne/vystupni praci si muzeme pomoci proudu
;; vytvorit pohodlnejsi rozhrani

;;; JEDNODUCHE VSTUPNI PROUDY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; vytvor stream ctenim vyrazu z vstupniho portu
(define input-port->stream
  (lambda (reader port)
    (let iter ()
      (let ((elem (reader port)))
        (if (eof-object? elem)
            (begin
              (close-input-port port)
              '())
            (cons-stream elem (iter)))))))

;; vytvori stream otevrenim souboru
(define file->stream
  (lambda (reader file-name)
    (input-port->stream reader (open-input-file file-name))))

;; Priklady: (cteni po vyrazech s pomoci ``read'' / po znacich ``read-char'')
(display-stream (file->stream read "data.txt"))
(display-stream (file->stream read-char "data.txt"))

;;; VYSTUPNI PROUDY
;;;;;;;;;;;;;;;;;;;

;; zapise stream do daneho otevreneho portu
(define stream->output-port
  (lambda (stream writer close? . port)
    (let ((port (if (null? port)
                    (current-output-port)
                    port)))
      (let iter ((stream stream))
        (if (null? stream)
            (if close? (close-output-port port))
            (begin
              (writer (stream-car stream) port)
              (iter (stream-cdr stream))))))))

; priklady pouziti:
(stream->output-port (stream 1 '(2 3) "4 5") write #f)
(stream->output-port (stream 1 '(2 3) "4 5") display #f)
(stream->output-port (stream 1 '(2 3) "4 5")
                     (lambda (x port)
                       (display x port)
                       (newline port))
                     #f)

;; otevre soubor a zapise do nej stream
(define stream->file
  (lambda (stream writer file-name)
    (let ((port (open-output-file file-name)))
      (stream->output-port steram writer #t port))))

;; INTERMEZZO: VYUZITI PROUDU KE ZVYSENI EFEKTIVNITY VYPOCTU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; maji dva seznamy stejne atomicke prvky pokud je projdeme zleva-doprava?
;; primocare reseni, ktere je neefektivni
(define equal-fringe?
  (lambda (s1 s2)

    ;; pomocna definice: linearizace seznamu
    (define flatten
      (lambda (l)
        (cond ((null? l) '())
              ((list? (car l))
               (append (flatten (car l))
                       (flatten (cdr l))))
              (else (cons (car l) (flatten (cdr l)))))))

    ;; jsou linearni seznamy totozne?
    (equal? (flatten s1) (flatten s2))))

(equal-fringe? '(a b c d) '(a b c d))           ; #t
(equal-fringe? '(a (b c) d) '((a b) c d))       ; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (d)))  ; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (e)))  ; #f
(equal-fringe? '(x (b (c)) () d) '(a b c (d)))  ; #f

;; Poznamka: v cem spociva neefektivita predchoziho reseni?
;;           1) konstruuji se linearni seznamy, ktere se pouzivaji jednorazove
;;           2) procedura se nechova ,,prirozene'' -- i presto, ze napriklad
;;              u seznamu (a ... (b ... je jasne ,,hned'', ze nebudou mit
;;              stejne atomy, procedura je presto cele prochazi

;; vylepseni efektivity procedury pomoci streamu pri zachovani
;; primocareho zapisu kodu

;; Reseni: budeme i dale linearizovat, ale misto linearlizace ,,do seznamu''
;;         budeme provadet linearizaci do streamu, ktere potom porovname:
;;         tim vyresime problem 2)

;; Potrebujeme dve nasledujici procedury:

;; predikat shodnosti dvou streamu
(define stream-equal?
  (lambda (s1 s2)
    (or (and (null? s1) (null? s2))
        (and (equal? (stream-car s1)
                     (stream-car s2))
             (stream-equal? (stream-cdr s1)
                            (stream-cdr s2))))))

; Pozn.: predchozi predikat potrebujeme, protoze dva prisliby jsou si rovny,
; p.k. je to tyz prislib, pr:
(define d (delay 1))
(equal? d d)                        ; #t
; ale:
(equal? (delay 1) (delay 1))        ; #f
; tim padem:
(equal? (stream 'a 'b 'c) (stream 'a 'b 'c))        ; #f ... nechceme
(stream-equal? (stream 'a 'b 'c) (stream 'a 'b 'c))        ; #t

;; tohe je ,,temer funkcni reseni''
(define equal-fringe?
  (lambda (s1 s2)

    (define called 0)

    ;; pomocna definice: linearizace seznamu
    (define flatten
      (lambda (l)
        ;; zobraz pocet volani, at vidime, ze to funguje
        (display "TICK: ")
        (set! called (+ 1 called))
        (display called)
        (newline)
        (cond ((null? l) '())
              ((list? (car l))
               (stream-append2 (flatten (car l))
                               (flatten (cdr l))))
              (else (cons-stream (car l)
                                 (flatten (cdr l)))))))

    ;; jsou linearni seznamy totozne?
    (stream-equal? (flatten s1) (flatten s2))))

;; kvuli zacyklenym seznamum je vhodne udelat strean-append2 jako formu!
;; tady je mensi zadrhel

;; tohle funguje + je videt, co to je efektivni
(equal-fringe? '(a b c d) '(a b c d))                ; #t
(equal-fringe? '(a (b c) d) '((a b) c d))        ; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (d)))        ; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (e)))        ; #f
(equal-fringe? '(x (b (c)) () d) '(a b c (d)))        ; #f

;; dalsi priklad, dva ,,zacyklene seznamy'', ktere evidentne nemaji st. listy
(define a '(a))
(define b '((b)))
(set-cdr! a a)
(set-cdr! b b)
; pri volani (equal-fringe? a b) predchozi procedura CYKLI!

;; Proc? Protoze pri volani stream-append2 jsou vyhodnoceny oba argumenty
;; (stream-append2 je procedura), takze nastane rekurzivni volani, ... ad inf.

;; reseni:

;; misto procedury stream-append2 vytvorime formu stream-append2
(define-macro stream-append2
  (lambda (s1 s2)
    `(let proc ((s ,s1))
       (if (stream-null? s)
           ,s2
           (cons-stream (stream-car s)
                        (proc (stream-cdr s)))))))

;; vsimnete si chovani:
(display-stream (stream-append2 (stream 1 2 3) (stream 'c 'd 'e)))
(stream-append2 (stream 1 2 3) ukrutna-blbost)        ; (1 . #<promise>)
; v predchozim pripade by PROCEDURA stream-append2 vratila chybu

;; novy pohled na makra:
;; ,,makra jsou procedury s line vyhodnocenymi argumenty''

;; nyni:
(define a '(a))
(define b '((b)))
(set-cdr! a a)
(set-cdr! b b)
; kvuli makroexpanzi je v nekterych interpretech potreba znova vyhodnotit
; equal-fringe?
; (equal-fringe? a b)        ; #f ... to jsme chteli (3 tiky)

;; 3. CAST: VYHODNOCOVACI PROCES S VYROVNAVACI PAMETI
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; tohle neni vedecky objev

;; cislicka
(define fib
  (lambda (n)
    (if (<= n 2)
        1
        (+ (fib (- n 1))
           (fib (- n 2))))))

(map fib '(1 2 3 4 5 6 7 8 9))

(fib 32)

;; iterativni verze predstavuje zrychleni, je vsak ,,mene citelna''
(define fib
  (lambda (n)
    (let iter ((a 1)
               (b 1)
               (n n))
      (if (<= n 1)
          a
          (iter b (+ a b) (- n 1))))))

(map fib '(1 2 3 4 5 6 7 8 9))

(fib 32)

;; urychleni vypoctu pomoci kappa-fka (Function with Known Arguments) vyrazu

;; prazdna pamet
(define empty-assoc
  (lambda () (cons (cons #f #f) '())))

;; destruktivne zarad novy zaznam nebo modifikuj existujici
;; procedura vraci nedefinovanou hodnotu
(define assoc!
  (lambda (assoc key val)
    (let iter ((as assoc))
      (cond ((null? as)
             (set-cdr! assoc
		       (cons (cons key val)
			     (cdr assoc))))
            ((equal? (caar as) key)
	     (set-cdr! (car as) val))
            (else (iter (cdr as)))))))

;; priklad zarazeni novych zaznamu:
(define a (empty-assoc))
a                    ; ((#f . #f))
(assoc! a 'ahoj 10)
a                    ; ((#f . #f) (ahoj . 10))
(assoc! a 'blah 20)
a                    ; ((#f . #f) (blah . 20) (ahoj . 10))
(assoc! a 'ahoj 30)
a                    ; ((#f . #f) (blah . 20) (ahoj . 30))

;; hledame pomoci assoc, protoze se jedna o asociativni seznam
(assoc 'blah a)  ; (blah . 20)
(assoc #f a)     ; (#f . #f)    ... tento par je vzdy pritomen kvuli mutaci

;; procedura memoize, ktera danou proceduru ``f'' zabali kodem, ktery udrzuje
;; hodnoty vysledku aplikace puvodni procedury ``f'' pro dane argumenty
(define memoize
  (lambda (f)
    (let ((memory (empty-assoc)))
      (lambda called-with-args
        (let ((found (assoc called-with-args memory)))
          (if found
              (cdr found)
              (let ((result (apply f called-with-args)))
                (assoc! memory called-with-args result)
                result)))))))

;; fibonacciho cisla urychlena pomoci memoize
(define fib
  (memoize 
   (lambda (n)
     (if (<= n 2)
         1
         (+ (fib (- n 1))
            (fib (- n 2)))))))

(fib 32)

;; POZOR: tohle by nefungovalo:

;; puvodni pomaly fib
(define fib
  (lambda (n)
    (if (<= n 2)
        1
        (+ (fib (- n 1))
           (fib (- n 2))))))

(define fast-fib (memoize fib))
(fast-fib 32)        ; bude ve skutecnosti pomaly, protoze se rekurzivne vola
                ; verze bez cacheovani vysledku, zapamatovana je pouze
                ; vysledna hodnota pro 32!

;; memoizovane procedury se chovaji jinak nez ,,klasicke'' v pripade
;; pouziti vedlejsiho efektu
(let ((cntr 0))
  (define modify (lambda (x) (set! cntr (+ 1 cntr))))
  (modify #f)
  (modify #f)
  cntr)                ; 2

(let ((cntr 0))
  (define modify (memoize (lambda (x) (set! cntr (+ 1 cntr)))))
  (modify #f)
  (modify #f)
  cntr)                ; 1

;; vylepsena verze memoize s limitovanou velikosti pamete
;; pokud je pamet prekrocena, je jeji obsah vysypan
(define make-memoize
  (lambda mem-limit
    (lambda (f)
      (let ((memory (empty-assoc)) (memory-size 0))
        (lambda called-with-args
          (let ((found (assoc called-with-args memory)))
            (if found
                (cdr found)
                (let ((result (apply f called-with-args)))
                  (if (and (not (null? mem-limit))
                           (> memory-size (car mem-limit)))
                      (begin
                        (set! memory-size 0)
                        (set! memory (empty-assoc))))
                  (assoc! memory called-with-args result)
                  (set! memory-size (+ memory-size 1))
                  result))))))))

(define fib
  ((make-memoize 5)
   (lambda (n)
     (if (<= n 2)
         1
         (+ (fib (- n 1))
            (fib (- n 2)))))))

(fib 32)

;; cim vetsi je buffer, tim je rychlejsi
;; srovnej (fib 32) pro mem-limit = 50, 20, 10, 5, 2, 1
;; treba pomoci globalnoho pocitadla:

(begin

(define count 0)
(define fib
  ((make-memoize 1)
   (lambda (n)
     (set! count (+ 1 count))
     (if (<= n 2)
         1
         (+ (fib (- n 1))
            (fib (- n 2)))))))
(fib 32)
count)

;; pro pohodlnejsi programovani procedur s vyrovnavaci pameti muzeme
;; novou specialni formu kappa:

;; globalni ,,konstanta'' udavajici velikost pameti
(define *max-memory* 1000)

;; spec. forma kappa
(define-macro kappa
  (lambda (args . body)
    `((make-memoize *max-memory*)
      (lambda ,args ,@body))))

; test:
(define fib
  (kappa (n)
    (if (<= n 2)
        1
        (+ (fib (- n 1))
           (fib (- n 2))))))

(fib 32)
(fib 500)        ; tohoto bychom se u klasicke
                ; rekurzivni procedury ,,nedockali''