
;;  Paradigmata Programovani 2B, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;; Ukazeme mechanismus ,,kontinuaci a unikovych funkci''
;; umoznuje a) vytvaret sirokou skalu (nelokalnich) skoku
;;          b) do jiste miry muzeme manipulovat s historii vypoctu jako s daty
;;             (napriklad budeme moci aktivovat vypocet, ktery jiz skoncil,
;;             tim, ze ,,skocime doprostred vypoctu'')

;; vse je zalozene na tom, ze v danem okamziku vypoctu jsme schopni zhmotnit
;; tzv. ,,kontext vypoctu'' a ,,unikovou funkci'' a dale je pouzivat

;; Scheme plne podporuje (zadny jiny PJ zatim plnohodnotne neobsahuje,
;; pouze nejake ,,nahrazky'',
;; POSIX obsahuje longjmp ... slaby odvar toho co ma Scheme)

;; 1. CAST: KONTEXT VYPOCTU, UNIKOVE FUNKCE, AKTUALNI POKRACOVANI
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; a) KONTEXT
;;;;;;;;;;;;;

;; neformalne: kontext je zhmotneni zbytku vypoctu, ktery by byl zahajen
;; okamzite po vyhodnoceni nejakeho podvyrazu

;; priklad:
(+ 2 (* 3 (/ 25 5)))

;; kontext (/ 25 5) je vypocet, ktery probehme po vyhodnoceni (/ 25 5)
;; v predchozim vyrazu, to jest hodnota se nasobi 3 a pak se k ni prida 2

;; kontext 3: hodnota je nasobena vysledkem vyhodnoceni (/ 25 5) a pak
;; je pridana dvojka

;; kontext *: pokud je hodnota funkce, pak se aplikuje s argumenty 3
;; a vysledkem vyhodnoceni (/ 25 5) a k tomu se jeste priste 2

;; program s dirou
;;;;;;;;;;;;;;;;;;

;; je de facto meta-program
;; dirou oznacujeme podvyraz, o jehoz kontext se zajimame

;; PROGRAM S DIROU: abstrakce: pohled na program, ve kterem mame misto
;; nektereho vyrazu uveden metasymbol \Hole (meta- protoze nepatri
;; do jazyka Scheme)

(+ 2 (* 3 (/ 25 5)))

(+ 2 (* 3 \Hole))

(+ 2 (* \Hole (/ 25 5)))

(+ 2 (\Hole 3 (/ 25 5)))

;; formalne: KONTEXT
;;;;;;;;;;;;;;;;;;;;

;; je procedura jednoho argumentu reprezentujici vypocet, ktery nastane po
;; dosazeni skutecne hodnoty misto diry do vyrazu

;; pro dany vyraz a vyse uvedene priklady s dirou
(+ 2 (* 3 (/ 25 5)))

;; si lze predstavit jako procedury vznikle vyhodnocenim nasledujicich vyrazu
(lambda (\Hole) (+ 2 (* 3 \Hole)))

(lambda (\Hole) (+ 2 (* \Hole (/ 25 5))))

(lambda (\Hole) (+ 2 (\Hole 3 (/ 25 5))))

;; b) UNIKOVE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;

;; globalni symbol na kterem budeme mit navazanou kontinuaci
(define *top-level-escaper* #f)

;; na *top-level-escaper* navaze dalsi pokracovani cyklu REPL
(call/cc
 (lambda (break)
   (set! *top-level-escaper* break)))

;; tohle je implementace ESCAPER pomoci CALL/CC
(define escaper
  (lambda (f)
    (lambda args
      (*top-level-escaper* (apply f args)))))

;; procedura se nazyva unikova procedura, pokud kdyz je aktivovana,
;; tak zprusobi okamzite preruseni vykonavani zbytku vypoctu a vysledkem
;; vyhodnoceni (celeho vstupniho vyrazu, ve kterem byla pouzita) je prave
;; vysledek jeji aplikace

;; unikove procedury budeme vytvaret pomoci konstruktu ESCAPER (neni v R5RS)

*		; procedura nasobeni
(escaper *)	; unikova procedura nasobeni

;; pouziti na top-level je stejne
(* 10 20)		; 200
((escaper *) 10 20)	; 200

;; avsak
(+ 2 (* 10 20))		   ; 202
(+ 2 ((escaper *) 10 20))  ; 200

;; dalsi priklad
(define p (escaper
	   (lambda (x)
	     (if (even? x)
		 x
		 (- x)))))

(p 10)
(p 11)

;; pomocna procedura pouzivajici predchozi `p'
(define test
  (lambda (x)
    (if (= 1 (p x))
	tohle-nikdy-neprobehne
	tohle-taky-ne)))

; ukazka toho, ze IF nikdy neprobehne
(test 10)              ; 10
(test 11)              ; 11
(eval `(test 10))      ; 10
(eval `(test (+ 5 6))) ; -11

;; POZNAMKA: unikova procedura je treba zabudovana procedura `error', ktera
;;           slouzi pro hlaseni chyb (pri vyhozeni chyby chceme okamzite
;;           prerusit zbytek vypoctu)

;; posleze si ukazene, ze v R5RS neni `escaper' zakladni koncept,
;; ale ze jej muzeme naprogramovat na par radku

;; c) AKTUALNI POKRACOVANI
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; procedura CALL/CC
;; (call/cc receiver)
;; receiver (prijemce) je procedura jednoho argumentu

;; pri (call/cc receiver) se provede
;; 1) vytvori se kontext (call/cc receiver) v aktualne vyhodnocovanem vyrazu
;; 2) receiver je zavolan s argumentem (ktery nazyvame aktualni pokracovani)
;;    jimz je procedura vznikla vyhodnocenim (escaper kontext)

;; takze pri pouziti (call/cc (lambda (f) ... bude uvnitr na f navazana
;; unikova procedura. Tuto unikovou proceduru muzeme aktivovat s jednim
;; argumentem. Pokud se tak stane, je okamzite preruseno vyhodnocovani
;; dalsiho kodu v receiveru a bude se pokracovat vyhodnocovanim kontextu
;; (call/cc (lambda (f) ... s hodnotou vznikou aplikaci f

(* 2 (+ 3 (/ 25 5)))	; 16

;; takze napriklad (f neni pouzito)
(* 2 (call/cc
      (lambda (f)
	(+ 3 (/ 25 5)))))	; 16

;; si lze predstavit:
(* 2 ((lambda (f)
	(+ 3 (/ 25 5)))
      (escaper (lambda (\Hole) (* 2 \Hole)))))

;; dalsi napriklad:
(* 2 (call/cc
      (lambda (f)
	(+ 3 (f 5)))))	; 10

;; si lze predstavit: (+ 3 ... se neuplatni, dojde k preruseni)
(* 2 ((lambda (f)
	(+ 3 (f 5)))
      (escaper (lambda (\Hole) (* 2 \Hole)))))	; 10

;; ted jiz stucne:

(+ 1 (call/cc (lambda (f) 2)))	; 3

(+ 1 (call/cc (lambda (f)
		(f 2))))	; 3

(+ 1 (call/cc (lambda (f)
		(if (even? (f 2))
		    100
		    200))))	; 3

(+ 1 (call/cc
      (lambda (f)
	(* 2 (call/cc
	      (lambda (g)
		(f (g 20))))))))	; 41

(+ 1 (call/cc
      (lambda (f)
	(* 2 (call/cc
	      (lambda (g)
		(g (f 20))))))))	; 21

;; 2. CAST: ESCAPING CONTINUATIONS (A JEJICH APLIKACE)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; termin: escaping continuation
;; kontinuace, ktera je pouzita k okamzitemu opusteni daneho vypoctu
;; dale ukazeme dalsi pouziti

;; PRIKLAD: okamzite opusteni rekurze
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; procedura realizujici soucin cisel v seznamu
(define list-product
  (lambda (l)
    (if (null? l)
	1
	(* (car l) (list-product (cdr l))))))

(list-product '(1 2 3 4 5))

;; muzeme pochopitelne vyjadrit:
(apply * '(1 2 3 4 5))

;; ale o to nam ted nejde
;; Otazka: jak zefektivnit list-product?
;; nasobeni nulou = vzdy nula

;; 1. reseni: pridame pouze podminku pro nulu
;; ma vyhodu: uz nezpracovavame prvky seznamu vyskytujici se za nulou
;; ma nevyhodu: faze odvijeni bude i nadale (zbytecne) probihat

;; rekurzivni procedura s pridanou podminkou
(define list-product
  (lambda (l)
    (cond ((null? l) 1)
	  ((= (car l) 0) 0)
	  (else (* (car l) (list-product (cdr l)))))))

;; 2. reseni: proceduru ponechame rekurzivni
;; aktualni pokracovani pouzijeme pro ukamzity unik z rekurze

;; rekurzivni verze pouzivajici call/cc
(define list-product
  (lambda (l)
    (call/cc
     (lambda (exit)
       (let proc ((l l))
	 (cond ((null? l) 1)
	       ((= (car l) 0) (exit 0))
	       (else (* (car l) (proc (cdr l))))))))))

;; dalsi variace, jde napsat napriklad takto:
(define list-product
  (lambda (l)
    (define proc
      (lambda (l return)
	(cond ((null? l) 1)
	      ((= (car l) 0) (return 0))
	      (else (* (car l) (proc (cdr l) return))))))
    (call/cc
     (lambda (f) (proc l f)))))

;; 3. reseni: prepiseme proceduru jako iterativni s akumulatorem,
;; iterativni proces lze snadno kdykoliv prerusit

;; iterativni verze
(define list-product
  (lambda (l)
    (let iter ((l l)
	       (accum 1))
      (if (null? l)
	  accum
	  (iter (cdr l) (* accum (car l)))))))

;; iterativni verze s okamzitym opustenim
(define list-product
  (lambda (l)
    (let iter ((l l)
	       (accum 1))
      (cond ((null? l) accum)
	    ((= (car l) 0) 0)
	    (else (iter (cdr l) (* accum (car l))))))))

;; Poznamka: v tomto pripade se muze zdat, ze call/cc je zbytecne
;; silny nastroj, problem jsme vyresili stejne dobre vytvorenim
;; iterativni verze procedury

;; PRIKLAD: opusteni hloubkove rekurze
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; V nekterych pripadech je ale vytvoreni iterativni verze procedury
;; dost slozite (kod se spatne cte) a proceduru je prirozene ponechat
;; rekurzivni

;; testuj, zda-li dane seznamy maji stejnou strukturu a jejich
;; atomy maji vzajemne dany vztah
(define atom-prop?
  (lambda (aprop? l1 l2)
    (cond ((and (null? l1) (null? l2)) #t)
	  ((and (pair? l1) (pair? l2))
	   (and (atom-prop? aprop? (car l1) (car l2))
		(atom-prop? aprop? (cdr l1) (cdr l2))))
	  ((and (not (pair? l1)) (not (pair? l2)))
	   (aprop? l1 l2))
	  (else #f))))

; priklady pouziti
(atom-prop? (lambda (x y) #t) '(1 (2 (3) 4)) '(1 (2 (3) 4)))	; #t
(atom-prop? (lambda (x y) #t) '(1 (2 ((3)) 4)) '(1 (2 (3) 4)))	; #f
(atom-prop? <= '(1 (2 (3) 4)) '(2 (3 (4) 5)))	; #t
(atom-prop? <= '(1 (2 (3) 4)) '(2 (3 (1) 5)))	; #f

;; nebo bez pomoci cond (jen s pouzitim `and' a `or')
(define atom-prop?
  (lambda (aprop? l1 l2)
    (or (and (null? l1)
	     (null? l2))
	(and (pair? l1)
	     (pair? l2)
	     (atom-prop? aprop? (car l1) (car l2))
	     (atom-prop? aprop? (cdr l1) (cdr l2)))
	(and (not (pair? l1))
	     (not (pair? l2))
	     (aprop? l1 l2)))))

;; ZEFEKTIVNENI: pokud kdykoliv korespondujici elementy nesplnuji danou
;; podminku nebo nesouhlasi jejich typ (jeden je par a druhy nikoliv)
;; zkytky stromu uz nemusime prohledavat (to zajistuje i predchozi reseni),
;; muzeme vsak rovnou ,,vyskocit z rekurze'' a vratit hodnotu #f

;; nebo bez pomoci cond (jen s pouzitim `and' a `or')
(define atom-prop?
  (lambda (aprop? l1 l2)
    (call/cc
     (lambda (exit)
       (let test ((l1 l1)
		  (l2 l2))
	 (or (and (null? l1)
		  (null? l2))
	     (and (pair? l1)
		  (pair? l2)
		  (test (car l1) (car l2))
		  (test (cdr l1) (cdr l2)))
	     (and (not (pair? l1))
		  (not (pair? l2))
		  (aprop? l1 l2))
	     (exit #f)))))))

;; predchzi proceduru lze prepsat iterativne, ale je to slozite
;; efektivita se ponekud vytraci, protoze rekurzivni volani procedury
;; nahrazujeme slozitou manipulaci se dvema dodatecnymi zasobniky
(define atom-prop?
  (lambda (aprop? l1 l2)
    (let test ((s1 '())	; pomocny zasobnik pro prvni seznam
	       (s2 '()) ; pomocny zasobnik pro druhy seznam
	       (l1 l1)
	       (l2 l2))
;(display (list s1 s2 l1 l2))
;(newline)
      (cond ((and (null? s1) (null? s2) (null? l1) (null? l1)) #t)
	    ((and (null? s1) (null? s2) (not (null? l1)) (not (null? l2)))
	     (test (cons (car l1) s1)
		   (cons (car l2) s2)
		   (cdr l1)
		   (cdr l2)))
	    ((and (null? s1) (null? s2)) #f)
	    ((or (null? s1) (null? s2)) #f)
	    ((and (null? (car s1)) (null? (car s2)))
	     (test (cdr s1) (cdr s2) l1 l2))
	    ((and (pair? (car s1)) (pair? (car s2)))
	     (test (cons (caar s1) (cons (cdar s1) (cdr s1)))
		 (cons (caar s2) (cons (cdar s2) (cdr s2)))
		 l1
		 l2))
	    ((or (pair? (car s1)) (pair? (car s2))) #f)
	    ((aprop? (car s1) (car s2))
	     (test (cdr s1) (cdr s2) l1 l2))
	    (else #f)))))

; priklady pouziti
(atom-prop? (lambda (x y) #t) '(1 (2 (3) 4)) '(1 (2 (3) 4)))	; #t
(atom-prop? (lambda (x y) #t) '(1 (2 ((3)) 4)) '(1 (2 (3) 4)))	; #f
(atom-prop? <= '(1 (2 (3) 4)) '(2 (3 (4) 5)))	; #t
(atom-prop? <= '(1 (2 (3) 4)) '(2 (3 (1) 5)))	; #f

;; 3. CAST: USCHOVA UNIKOVYCH FUNKCI A JEJICH POZDEJSI POUZITI
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; kontinuace jsou prcedury
(call/cc (lambda (f) f))              ; #<continuation>
(procedure? (call/cc (lambda (f) f))) ; #t

;; tj. kontinuace = element prvniho radu
(define blah #f)

;; receiver v nasledujicim vyrazu nejprve navaze
;; kontinuaci na globalni symbol `blah'
(* 2 (call/cc
      (lambda (f)
	(set! blah f)
	30)))		; 60

;; kontinuaci v `blah' je mozne pouzivat i kdyz puvodni ,,vypocet uz skoncil''
(blah 1)	; 2
(blah 2)	; 4
(blah 3)	; 6

;; dalsi priklad
(begin
  (display "JEDNOU")
  (newline)
  (* 2 (call/cc
	(lambda (f)
	  (set! blah f)
	  30))))

(blah 10)	; 20 (nic se netiskne)

;; dalsi priklad
(* (let ((result (call/cc
		  (lambda (f)
		    (set! blah f)
		    30))))
     (display "POKAZDE")
     (newline)
     result))	; 30

(blah 10)	; 20 (tiskne se `POKAZDE')

;; zajimavy efekt
(* 2 (call/cc
      (lambda (f)
	(set! blah f)
	zde-udelame-zamerne-chybu)))  ; error

;; chyba nastala az po navazani `blah', takze:
(blah 10)	; 20 (je v poradku)

;; PRIKLAD: ESCAPER pomoci CALL/CC
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; globalni symbol na kterem budeme mit navazanou kontinuaci
(define *top-level-escaper* #f)

;; na *top-level-escaper* navaze dalsi pokracovani cyklu REPL
(call/cc
 (lambda (break)
   (set! *top-level-escaper* break)))

;; tohle je implementace ESCAPER pomoci CALL/CC
(define escaper
  (lambda (f)
    (lambda args
      (*top-level-escaper* (apply f args)))))

;; PRIKLAD: osetrovani chyb
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; globalni symbol na kterem budeme mit navazanou kontinuaci
(define *top-level-escaper* #f)

;; na *top-level-escaper* navaze dalsi pokracovani cyklu REPL
(call/cc
 (lambda (break)
   (set! *top-level-escaper* break)))

;; implementace chyboveho hlaseni 
;; po zavolani error, je zobrazena hlaska a omakzite preruseno
;; vykonavani aktualniho kodu, interpret ceka na zadani nalsiho vyrazu
(define error
  (lambda (proc message . values)
    (display "Error {")
    (display (symbol->string proc))
    (display "}: ")
    (display message)
    (if (not (null? values))
	(begin
	  (display " {")
	  (let ((first #t))
	    (for-each
	     (lambda (x)
	       (if first
		   (set! first #f)
		   (display ", "))
	       (display x))
	     values))
	  (display "}")))
    (newline)
    (*top-level-escaper*)))

;; priklad pouziti
(define list-product
  (lambda (l)
    (call/cc
     (lambda (exit)
       (let proc ((l l))
	 (cond ((null? l) 1)
	       ((not (pair? l))
		(error 'list-product
		       "Given argument is not a list" l))
	       ((not (number? (car l)))
		(error 'list-product
		       "List member is not a number" (car l)))
	       ((= (car l) 0) (exit 0))
	       (else (* (car l) (proc (cdr l))))))))))

(list-product '(1 2 3 4 5))
(list-product '(1 2 3 . 4))
(list-product '(1 2 blah 4))

;; PRIKLAD: system vyjimek
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; vyjimky jsou pokrocilejsi zpusob osetrovani vyjimecnych (chybovych)
;; situaci, ktere mohou nastat behem cinnosti programu

;; Pouziti:
;; (catch nazev-vyjimky telo...)
;; uvnitr tela je mozne pouzit (throw nazev-nejake-vyjimky vyraz)

;; top-level procedura `throw'
(define throw
  (lambda (exception-name expr)
    (error 'throw "Top level throw triggered with" exception-name)))

;; makro `catch'
(define-macro catch
  (lambda (label . prgn)
    (let ((exit (gensym)))
      `(call/cc
	(lambda (,exit)
	  (let ((throw (lambda (exception-name expr)
			 (if (eq? ,label exception-name)
			     (,exit expr)
			     (throw exception-name expr)))))
	    (begin ,@prgn)))))))

; priklad pouziti
(catch 'chyba
  (+ 1 2)
  (* 3 4))

; priklad pouziti
(catch 'chyba
  (+ 1 2)
  (throw 'chyba 20)
  (* 3 4))

; priklad pouziti
(catch 'chyba
  (+ 1 2)
  (throw 'neexistujici-vyjimka 20)
  (* 3 4))

; priklad pouziti
(catch 'chyba
  (display "X")
  (newline)
  (catch 'dalsi
    (+ 1 2)
    (display "Y")
    (newline)
    (throw 'dalsi 20)
    (* 3 4))
  (display "Z")
  (newline))		; vypise se X, Y i Z

; priklad pouziti
(catch 'chyba
  (display "X")
  (newline)
  (catch 'dalsi
    (+ 1 2)
    (display "Y")
    (newline)
    (throw 'chyba 20)
    (* 3 4))
  (display "Z")
  (newline))		; v tomto pripade se "Z" nevypise

; dalsi priklad
(define f
  (lambda (n)
    (let ((x 10))
      (catch 'bar
             (catch 'foo
                    (if (> n 0)
                        (throw 'foo (set! x 40)))
                    (if (< n 0)
                        (throw 'bar (+ 1 2)))
                    (set! x 100))
	     (* 2 x)))))

(f -1) ; 3
(f 0)  ; 200
(f 1)  ; 80

;; PRIKLAD: aserce a testovani robustnosti programu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; toto je priklad na neprime vyuziti kontinuaci

;; aserce, pouziti top-level vyskoku
(define-macro assert
  (lambda (proc-name assertions . body)
    `(cond ,@(map (lambda (ass)
		    `((not ,(car ass))
		      (begin
			(display "ASSERT {")
			(display ,proc-name)
			(display "}: ")
			,@(map (lambda (x)
				 `(display ,x))
			       (cdr ass))
			(newline)
			(*top-level-escaper*))))
		  assertions)
	    (else (begin ,@body)))))

;; Priklad: faktorial s asercemi
(define fak
  (lambda (n)

    ;; aserce pro proceduru faktorial
    (assert "faktorial"
      (((number? n) "argument supposed to be a number, given: " n)
       ((exact? n) n " must be an exact number")
       ((>= n 0) n " must be a non-negative integer"))

      ;; vlastni telo procedury
      (if (= n 0)
	  1
	  (* n (fak (- n 1)))))))

(fak 6)
(fak 'blah)
(fak 1.2)
(fak -2)

;; pri finalni verzi programu (odladene), je mozne aserce ignorovat
;; nasledujici redefinici makra, kod vsech procedur pritom zustava
;; nadale stejny, vyhodne treba pri kompilaci (Bigloo)

(define-macro assert
  (lambda (proc-name assertions . body)
    `(begin ,@body)))

;; PRIKLAD: jednoduchy debugger
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; globalni symboly na kterem jsou navzany break-point a zastaveni
(define *next* #f)
(define *stop* #f)

;; nastaveni stopky
(call/cc
 (lambda (f)
   (set! *stop* f)))

;; prerus vypocet
(define stop-execution
  (lambda ()
    (*stop*)))

;; zapis do *next* aktualni pokracovani v danem bode
(define break-point
  (lambda ()
    (call/cc
     (lambda (f)
       (set! *next* f)
       (stop-execution)
       hic-sunt-leones))))

;; pokracuj ve vypoctu
(define run
  (lambda ()
    (*next*)))

;; fibonacciho cisla (rekurzivni varianta)
(define fib
  (lambda (n)
    (display "CALL: ")
    (display n)
    (newline)
    (break-point)
    (let ((return (if (<= n 1)
		      1
		      (+ (fib (- n 2)) (fib (- n 1))))))
      (display "Return from: ")
      (display n)
      (display ", Value: ")
      (display return)
      (newline)
      (break-point)
      return)))

(fib 6)
(run)
(run)
(run)
(run)

;; 4. CAST: ITERATORY A JEJICH APLIKACE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; PRIKLAD: ITERATORY NA LINEARNICH SEZNAMECH
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; identifikator ukonceni iterace
(define *end-of-iteration* (lambda () #f))

;; predikat indikujici konec iterace
(define finished?
  (lambda (elem)
    (eq? elem *end-of-iteration*)))

;; pro dany seznam vygeneruje iterator
(define generate-iterator
  (lambda (l)
    (letrec ((return #f)
	     (start
	      (lambda ()
		(let loop ((l l))
		  (if (null? l)
		      (return *end-of-iteration*)
		      (begin
			(call/cc
			 (lambda (new-start)
			   (set! start new-start)
			   (return (car l))))
			(loop (cdr l))))))))
      (lambda ()
	(call/cc
	 (lambda (f)
	   (set! return f)
	   (start)))))))

; priklad pouziti:
(define p (generate-iterator '(a b c d e)))
(p)
(p)
(p)
(p)
(p)
(p)
(eq? (p) *end-of-iteration*)

;; PRIKLAD: ITERATORY NA HLOUBKOVYCH SEZNAMECH
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pro dany seznam vygeneruje hloubkovy iterator
(define generate-depth-iterator
  (lambda (l)
    (letrec ((return #f)
	     (start
	      (lambda ()
		(let loop ((l l))
		  (cond ((null? l) 'nejaka-hodnota)
			((pair? l)
			 (begin
			   (loop (car l))
			   (loop (cdr l))))
			(else
			 (call/cc
			  (lambda (new-start)
			    (set! start new-start)
			    (return l))))))
		(return '()))))
      (lambda ()
	(call/cc
	 (lambda (f)
	   (set! return f)
	   (start)))))))

; () staci na detekci konce, protoze jdeme pres atomy

; priklad pouziti:
(define p (generate-depth-iterator '(a (b (c (d)) e))))
(define q (generate-depth-iterator '(((a b) ((c d))) e)))
(p)
(q)
(p)
(q)
(p)
(q)
(p)
(q)
(p)
(q)
(p)
(q)

;; INTERMEZZO: VYUZITI CALL/CC KE ZVYSENI EFEKTIVNITY VYPOCTU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; ukazka pouzitiiteratoru

;; OPET SI VEZMEME PROCEDURU:

;; maji dva seznamy stejne atomicke prvky pokud je projdeme zleva-doprava?
;; primocare reseni, ktere je neefektivni
(define equal-fringe?
  (lambda (s1 s2)

    ;; pomocna definice: linearizace seznamu
    (define flatten
      (lambda (l)
	(cond ((null? l) '())
	      ((list? (car l)) (append (flatten (car l)) (flatten (cdr l))))
	      (else (cons (car l) (flatten (cdr l)))))))

    ;; jsou linearni seznamy totozne?
    (equal? (flatten s1) (flatten s2))))

;; na prednasce o linem vyhodnocovani jsme ukazali jednu z metod jejiho
;; zefektivneni, pripominam:

;; Poznamka: v cem spociva neefektivita predchoziho reseni?
;;           1) konstruuji se linearni seznamy, ktere se pouzivaji jednorazove
;;           2) procedura se nechova ,,prirozene'' -- i presto, ze napriklad
;;              u seznamu (a ... (b ... je jasne ,,hned'', ze nebudou mit
;;              stejne atomy, procedura je presto cele prochazi

;; Bod 2) uz umime vyresit treba pomoci proudu.
;; pomoci aktualniho pokracovani ted vyresime i bod 1):
;; budeme mit k dispozici proceduru, ktera bude schopna testovat shodu
;; atomickych prvku seznamu bez konstrukce jedineho (noveho) paru

;; PRIKLAD: RESENI PROBLEMU
;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; nova, efektivni verze equal-fringe?
(define equal-fringe?
  (lambda (s1 s2)
    (let ((iter1 (generate-depth-iterator s1))
	  (iter2 (generate-depth-iterator s2)))
      (let test ((v1 (iter1))
		 (v2 (iter2)))
	(or (and (null? v1) (null? v2))
	    (and (equal? v1 v2) (test (iter1) (iter2))))))))

(equal-fringe? '(a b c d) '(a b c d))		; #t
(equal-fringe? '(a (b c) d) '((a b) c d))	; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (d)))	; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (e)))	; #f
(equal-fringe? '(x (b (c)) () d) '(a b c (d)))	; #f
