
;;  SCHEME IN SCHEME (STACK-BASED IMPLEMENTATION)
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;;
;;;;  SYSTEM MANIFESTOVANYCH TYPU
;;;;

;; vytvor element jazyka s manifestovanym typem
(define curry-make-elem
  (lambda (type-tag)
    (lambda (data)
      (cons type-tag data))))

;; vrat visacku s typem
(define get-type-tag car)

;; vrat data
(define get-data cdr)

;; test daneho typu
(define curry-scm-type
  (lambda (type)
    (lambda (elem)
      (equal? type (get-type-tag elem)))))

;;;;
;;;;  ZAKLADNI ELEMENTY JAZYKA SCHEME
;;;;

;;;; CISLA
;;;;;;;;;;;;;;;;;;;;

;; konstruktor cisla a test zda-li se jedna o cislo
(define make-number (curry-make-elem 'number))
(define scm-number? (curry-scm-type 'number))

;;;; SYMBOLY
;;;;;;;;;;;;;;;;;;;;

;; konstruktor symbolu a test zda-li se jedna o symbol
(define make-symbol (curry-make-elem 'symbol))
(define scm-named-symbol? (curry-scm-type 'symbol))

;; konstruktor generovaneho symbolu
(define make-generated-symbol
  (let ((make-physical-element (curry-make-elem 'generated-symbol)))
    (lambda ()
      (make-physical-element (cons #f #f)))))
(define scm-generated-symbol? (curry-scm-type 'generated-symbol))

;; test zda-li se jedna o symbol pojmenovany/generovany
(define scm-symbol?
  (lambda (elem)
    (or (scm-named-symbol? elem)
	(scm-generated-symbol? elem))))

;;;; TECKOVE PARY 
;;;;;;;;;;;;;;;;;;;;

;; konstruktor paru CONS
(define make-pair
  (let ((make-physical-pair (curry-make-elem 'pair)))
    (lambda (head tail)
      (make-physical-pair (cons head tail)))))

;; test datoveho typu
(define scm-pair? (curry-scm-type 'pair))

;; selektor paru CAR
(define pair-car
  (lambda (pair)
    (if (scm-pair? pair)
	(car (get-data pair))
	(error "; Car: argument must be a pair"))))

;; selektor paru CDR
(define pair-cdr
  (lambda (pair)
    (if (scm-pair? pair)
	(cdr (get-data pair))
	(error "; Cdr: argument must be a pair"))))

;; mutator paru SET-CAR!
(define pair-set-car!
  (lambda (pair value)
    (if (scm-pair? pair)
	(begin
	  (set-car! (get-data pair) value)
	  the-void-value)
	(error "; Set-car!: argument must be a pair"))))

;; mutator paru SET-CDR!
(define pair-set-cdr!
  (lambda (pair value)
    (if (scm-pair? pair)
	(begin
	  (set-cdr! (get-data pair) value)
	  the-void-value)
	(error "; Set-cdr!: argument must be a pair"))))

;; obecny konvertor seznamu na scm-seznam a obracene
(define convert-list
  (lambda (a-null? a-car a-cdr b-cons b-nil f l)
    (if (a-null? l)
	b-nil
	(b-cons (f (a-car l))
		(convert-list a-null? a-car a-cdr b-cons b-nil f
			      (a-cdr l))))))

;;;; PROSTREDI
;;;;;;;;;;;;;;;;;;;;;;;;;

;; preved asociacni seznam na tabulku prostredi
(define assoc->env
  (lambda (l)
    (if (null? l)
	the-empty-list
	(make-pair (make-pair
		    (make-symbol (caar l))
		    (cdar l))
		   (assoc->env (cdr l))))))

;; konstruktor prostredi
(define make-env
  (let ((make-physical-env (curry-make-elem 'environment)))
    (lambda (pred table)
      (make-physical-env
       (cons pred (make-pair scm-false table))))))

;; test datoveho typu
(define scm-env? (curry-scm-type 'environment))

;; konstruktor globalniho prostredi
(define make-global-env
  (lambda (alist-table)
    (make-env scm-false (assoc->env alist-table))))

;; vrat tabulku
(define get-table
  (lambda (elem)
    (if (scm-env? elem)
	(cdr (get-data elem))
	(error "; Get-table: argument must be an environment"))))

;; vrat predka
(define get-pred
  (lambda (elem)
    (if (scm-env? elem)
	(car (get-data elem))
	(error "; Get-pred: argument must be an environment"))))

;; je globalni prostredi?
(define global?
  (lambda (elem)
    (and (scm-env? elem)
	 (equal? scm-false (get-pred elem)))))

;; hledani vazeb v asociacnim poli
(define scm-assoc
  (lambda (key alist)
    (cond ((scm-null? alist) scm-false)
	  ((eq? key (pair-car (pair-car alist))) (pair-car alist))
	  ((equal? key (pair-car (pair-car alist))) (pair-car alist))
	  (else (scm-assoc key (pair-cdr alist))))))

;; vyhledej vazbu v prostredi env, nebo vrat not-found
(define lookup-env
  (lambda (env symbol search-nonlocal? not-found)
    (let ((found (scm-assoc symbol (pair-cdr (get-table env)))))
      (cond ((not (equal? found scm-false)) found)
	    ((global? env) not-found)
	    ((not search-nonlocal?) not-found)
	    (else (lookup-env (get-pred env) symbol #t not-found))))))

;;;; PRIMITIVNI PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor syrove primitivni procedury (ma kontrolu nad zasobniky)
(define make-manipulator (curry-make-elem 'manipulator))
(define scm-manipulator? (curry-scm-type 'manipulator))

;; konstruktor primitivni procedury a test zda-li se jedna o proceduru
(define make-primitive (curry-make-elem 'primitive))
(define scm-primitive? (curry-scm-type 'primitive))

;; vytvareni primitivnich procedur pomoci wrapperu
(define wrap-primitive
  (lambda (proc)
    (make-primitive
     (lambda arguments
       (expr->intern (apply proc (map get-data arguments)))))))

;; vytvareni predikatu pro testy typu pomoci wrapperu
(define wrap-predicate
  (lambda (type-pred)
    (make-primitive
     (lambda elems
       (if (apply type-pred elems)
	   scm-true
	   scm-false)))))

;;;; UZIVATELSKE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor procedury
(define make-procedure
  (let ((make-physical-procedure (curry-make-elem 'procedure)))
    (lambda (env args body)
      (make-physical-procedure (list env args body)))))

;; selektory
(define procedure-environment (lambda (proc) (car (get-data proc))))
(define procedure-arguments (lambda (proc) (cadr (get-data proc))))
(define procedure-body (lambda (proc) (caddr (get-data proc))))

;; test zda-li se jedna o proceduru
(define scm-user-procedure? (curry-scm-type 'procedure))

;; test zda-li se jedna o primitivni/uzivatelskou proceduru
(define scm-procedure?
  (lambda (elem)
    (or (scm-primitive? elem)
	(scm-user-procedure? elem)
	(and (scm-manipulator? elem)
	     (scm-procedure? (get-data elem))))))

;;;; PRIMITIVNI SPECIALNI FORMY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor primitivni spec. formy a test zda-li se jedna o spec. formu
(define make-specform (curry-make-elem 'specform))
(define scm-specform? (curry-scm-type 'specform))

;;;; MAKRA
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstrukto makra
(define make-macro (curry-make-elem 'macro))
(define scm-macro? (curry-scm-type 'macro))

;; test zda-li se jedna o primitivni/uzivatelskou formu
(define scm-form?
  (lambda (elem)
    (or (scm-specform? elem)
	(scm-macro? elem)
	(and (scm-manipulator? elem)
	     (scm-form? (get-data elem))))))

;;;; SPECIALNI ELEMENTY JAZYKA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pravdivostni hodnoty
(define scm-false ((curry-make-elem 'boolean) #f))
(define scm-true  ((curry-make-elem 'boolean) #t))
(define scm-boolean? (curry-scm-type 'boolean))

;; prazdny seznam
(define the-empty-list ((curry-make-elem 'empty-list) '()))
(define scm-null? (lambda (elem) (eq? elem the-empty-list)))

;; vrazdna hodnota
(define the-void-value ((curry-make-elem 'void) '()))
(define scm-void? (lambda (elem) (equal? elem the-void-value)))

;; nedefinovana hodnota
(define the-undefined-value ((curry-make-elem 'undefined) '()))
(define scm-undefined? (lambda (elem) (equal? elem the-undefined-value)))

;; specialni elementy ridici vyhodnocovani
(define the-inspect ((curry-make-elem 'stack-indicator-inspect) '()))
(define the-envrun ((curry-make-elem 'stack-indicator-environment-run) '()))
(define the-run ((curry-make-elem 'stack-indicator-run) '()))
(define the-return ((curry-make-elem 'stack-indicator-return) '()))
(define the-rotrue ((curry-make-elem 'stack-indicator-return-on-true) '()))
(define the-discard ((curry-make-elem 'stack-indicator-discard) '()))
(define the-defasgn ((curry-make-elem 'stack-indicator-define-assign) '()))
(define the-setasgn ((curry-make-elem 'stack-indicator-set-assign) '()))

;;;; GENERICKE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; scm-eq? naprogramujeme pomoci interniho eq?
(define scm-eq?
  (lambda (elem-a elem-b)
    (eq? (get-data elem-a)
	 (get-data elem-b))))


;;;;
;;;;  READER (nacteni vyrazu a prevedeni do interni representace)
;;;;

;; prevedeni vyrazu do interni formy
(define expr->intern
  (lambda (expr)
    (cond ((symbol? expr) (make-symbol expr))
	  ((number? expr) (make-number expr))
	  ((and (boolean? expr) expr) scm-true)
	  ((boolean? expr) scm-false)
	  ((null? expr) the-empty-list)
	  ((pair? expr) (make-pair (expr->intern (car expr))
				   (expr->intern (cdr expr))))
	  ((eof-object? expr) #f)
	  (else (error "; Syntactic error.")))))

;; nacti vstupni vyraz do interni formy
(define scm-read
  (lambda ()
    (expr->intern (read))))

;;;;
;;;;  PRINTER (vytisteni vyrazu v jeho externi reprez. -- pokud je to mozne)
;;;;

;; pouze pouzije display -- vypise syrovou reprezentaci
(define scm-print
  (lambda (elem)
    (display elem)))

;; prevedeni elementu do externi formy
(define intern->expr
  (lambda (expr)
    (cond ((scm-pair? expr) (cons (intern->expr (pair-car expr))
				  (intern->expr (pair-cdr expr))))
	  ((scm-env? expr) "#<environment>")
	  ((scm-primitive? expr) "#<primitive-procedure>")
	  ((scm-user-procedure? expr) "#<user-defined-procedure>")
	  ((scm-macro? expr) "#<macro>")
	  ((scm-manipulator? expr)
	   (string-append "#<manipulator:"
			  (intern->expr (get-data expr))
			  ">"))
	  ((scm-generated-symbol? expr) "#<generated-symbol>")
	  ((scm-specform? expr) "#<special-form>")
	  ((scm-undefined? expr) "#<undefined>")
	  ((scm-void? expr) "#<void>")
	  ((eq? expr the-inspect) "#<inspect>")
	  ((eq? expr the-run) "#<run>")
	  ((eq? expr the-return) "#<return>")
	  ((eq? expr the-rotrue) "#<return-on-true>")
	  ((eq? expr the-discard) "#<discard>")
	  ((eq? expr the-defasgn) "#<define-assign>")
	  ((eq? expr the-setasgn) "#<set-assign>")
	  (else (get-data expr)))))

;; o neco ,,chytrejsi printer''
(define scm-print
  (lambda (elem)
    (display (intern->expr elem))))

;; vytisteni obsahu zasobniku ridicich vypocet
(define print-stack
  (lambda (stack title)
    (display "; ")
    (display title)
    (display ": ")
    (let next ((l stack))
      (if (not (null? l))
	  (if (null? (cdr l))
	      (scm-print (car l))
	      (begin
		(scm-print (car l))
		(display ", ")
		(next (cdr l))))))
    (display "]\n")))

;;;;
;;;;  PROCEDURY PRO PREVOD DATOVE REPREZENTACE
;;;;

;; map pres elementy tvorici scm-seznam, vysledkem je klasicky seznam
(define map-scm-list->list
  (lambda (f scm-l)
    (convert-list scm-null? pair-car pair-cdr cons '() f scm-l)))

;; preved scm-seznam na klasicky seznam
(define scm-list->list
  (lambda (scm-l)
    (convert-list scm-null? pair-car pair-cdr cons '() (lambda (x) x) scm-l)))

;; preved klasicky seznam na scm-seznam
(define list->scm-list
  (lambda (l)
    (convert-list null? car cdr make-pair the-empty-list (lambda (x) x) l)))

;; list-head: vrat prvnich n prvku z daneho seznamu
(define list-head
  (lambda (list n)
    (if (<= n 0)
	'()
	(cons (car list)
	      (list-head (cdr list) (- n 1))))))

;; vytvor tabulku vazeb: formalni argument -- argument
(define make-bindings
  (lambda (formal-args args)
    (cond ((scm-null? formal-args) the-empty-list)
	  ((scm-symbol? formal-args)
	   (make-pair (make-pair formal-args (list->scm-list args))
		      the-empty-list))
	  (else (make-pair
		 (make-pair (pair-car formal-args) (car args))
		 (make-bindings (pair-cdr formal-args) (cdr args)))))))

;;;;
;;;;  PRAVIDLA PRO ZASOBNIKOVY EVALUATOR
;;;;

;;;; GLOBALNI ZASOBNIKY RIDICI VYHODNOCOVANI

(define *EXEC* '())    ;; zasobnik vstupnich vyrazu
(define *RSLT* '())    ;; zasobnik vysledku

;;;; ZPRACOVANI SYMBOLU

;; je na vrcholu *EXEC* symbol?
(define rule-eval-symbol?
  (lambda (*EXEC* *RSLT*)
    (and (scm-symbol? (car *EXEC*))
	 (scm-env? (cadr *EXEC*)))))

;; smaz symbol z vrcholu *EXEC* a na vrchol *RSLT* zapis vazbu symbolu
(define rule-eval-symbol
  (lambda ()
    (set! *RSLT*
	  (cons
	   (let* ((binding (lookup-env (cadr *EXEC*) (car *EXEC*) #t #f)))
	     (if binding
		 (pair-cdr binding)
		 (error "; EVAL: Symbol not bound")))
	   *RSLT*))
    (set! *EXEC* (cddr *EXEC*))))

;;;; ZPRACOVANI SEZNAMU

;; je na vrcholu *EXEC* seznam?
(define rule-eval-list?
  (lambda (*EXEC* *RSLT*)
    (scm-pair? (car *EXEC*))))

;; na vrcholu *EXEC* je seznam:
;; na vrchol se vlozi prvni prvek seznamu na vrcholu *EXEC*
;; pripravi se `inspect'
(define rule-eval-list
  (lambda ()
    (set! *EXEC*
	  `(,(pair-car (car *EXEC*)) ; prvni prvek seznamu na vrchlolu *EXEC*
	    ,(cadr *EXEC*)           ; prostredi
	    ,the-inspect             ; specialni element `inspect'
	    ,(pair-cdr (car *EXEC*)) ; seznam z vrcholu *EXEC* bez 1. prvku
	    ,@(cdr *EXEC*)))))       ; zbytek *EXEC* (zacina prostredim)

;;;; INSPEKCE VYHODNOCENI PRVNIHO PRVKU SEZNAMU

;; je na vrcholu *EXEC*  specialni element `the-inspect' a *RSLT* je neprazdny?
(define rule-eval-the-inspect?
  (lambda (*EXEC* *RSLT*)
    (and (eq? (car *EXEC*) the-inspect)
	 (not (null? *RSLT*)))))

;; na vrcholu *EXEC* je `the-inspect'
;; na vrcholu *RSLT* je element vznikly vyhodnocenim prvniho prvku seznamu
;; rozlisuleme situaci podle toho, zda-li je prvek na vrcholu *RSLT*
;; a) procedura (prmitivni nebo uzivatelska), b) spec. forma nebo makro
(define rule-eval-the-inspect
  (lambda ()
    (let* ((args (scm-list->list (cadr *EXEC*)))
	   (nargs (make-number (length args)))
	   (env (caddr *EXEC*)))
      (cond 

       ;; prvni argument se vyhodnotil na proceduru
       ((scm-procedure? (car *RSLT*))
	(set! *EXEC*
	      `(,@(apply append
			 (map (lambda (x)
				`(,x ,env))
			      args)) ; priprav na vyhodnoceni zbylych arg.
		,the-run             ; specialni element `run'
		,nargs               ; zaznam o pocet aktualnich argumentu
		,env                 ; prostredi
		,@(cdddr *EXEC*))))  ; zbytek seznamu *EXEC*

       ;; prvni element se vyhodnotil na spec. formu / makro
       ((scm-form? (car *RSLT*))
	(begin
	  (set! *EXEC*
		`(,the-run           ; specialni element `run'
		  ,nargs             ; zaznam o pocet aktualnich argumentu
		  ,env               ; prostredi
		  ,@(cdddr *EXEC*))) ; zbytek seznamu *EXEC*
	  (set! *RSLT*
		(append args *RSLT*))))

       ;; prvni element se vyhodnotil na neco necekaneho
       (else (error "; INS: 1st elem. did not eval. to proc./form."))))))

;;;; VYHODNOCOVANI NA ZAHLADE PREDCHOZI INSPEKCE

;; je na vrcholu *EXEC* specialni element `the-run'?
(define rule-eval-the-run?
  (lambda (*EXEC* *RSLT*)
    (eq? (car *EXEC*) the-run)))

;; na vrcholu *EXEC* je `run', provede se aplikace procedury
(define rule-eval-the-run
  (lambda ()
    (let* ((nargs (get-data (cadr *EXEC*)))  ; pocet argumentu
	   (env (caddr *EXEC*))              ; prostredi aplikace
	   (proc (list-ref *RSLT* nargs))    ; procedura
	   (args (list-head *RSLT* nargs)))  ; seznam argumentu
      (cond

       ;; aplikace primitivni procedury
       ((scm-primitive? proc)
	(begin
	  (set! *RSLT*
		`(,(apply (get-data proc)
			  (reverse args))            ; vloz vysledek aplikace
		  ,@(list-tail *RSLT* (+ 1 nargs)))) ; smaz argumenty z *RSLT*
	  (set! *EXEC* (cdddr *EXEC*))))             ; zkrat *EXEC*

       ;; aplikace uzivatelske procedury
       ((scm-user-procedure? proc)
	(begin
	  (set! *EXEC*
		`(,(procedure-body proc)
		  ,(make-env (procedure-environment proc)
			     (make-bindings (procedure-arguments proc)
					    (reverse args)))
		  ,@(cdddr *EXEC*)))
	  (set! *RSLT* (list-tail *RSLT* (+ 1 nargs)))))

       ;; aplikace specialni formy
       ((scm-specform? proc)
	(begin
	  (set! *RSLT*
		`(,(apply (get-data proc) env args)  ; vloz vysledek aplikace
		  ,@(list-tail *RSLT* (+ 1 nargs)))) ; smaz argumenty z *RSLT*
	  (set! *EXEC* (cdddr *EXEC*))))             ; zkrat *EXEC*

       ;; aplikace makra
       ((scm-macro? proc)
	(begin
	  (set! *RSLT*
		`(,@(reverse (list-head *RSLT* nargs))
		  ,(get-data proc)
		  ,@(list-tail *RSLT* (+ 1 nargs))))
	  (set! *EXEC*
		`(,@(list-head *EXEC* 3)
		  ,the-return
		  ,(caddr *EXEC*)
		  ,@(list-tail *EXEC* 3)))))

       ;; aplikace manipulatoru (primitivni procedury)
       ((and (scm-manipulator? proc)
	     (scm-primitive? (get-data proc)))
	(let ((new-stacks ((get-data (get-data proc)) *EXEC* *RSLT*)))
	  (set! *EXEC* (car new-stacks))
	  (set! *RSLT* (cadr new-stacks))))

       ;; aplikace manipulatoru (spec. formy)
       ((and (scm-manipulator? proc)
	     (scm-specform? (get-data proc)))
	(let ((new-stacks ((get-data (get-data proc))
			   *EXEC* *RSLT*)))
	  (set! *EXEC* (car new-stacks))
	  (set! *RSLT* (cadr new-stacks))))

       ;; v ostatnich pripadech dojde k chybe
       (else (error "; RUN: cannot apply"))))))

;;;; VYHODNOCOVANI (SE ZADANYM LEX. PREDKEM) NA ZAHLADE PREDCHOZI INSPEKCE

;; je na vrcholu *EXEC* specialni element `the-envrun'?
(define rule-eval-the-envrun?
  (lambda (*EXEC* *RSLT*)
    (eq? (car *EXEC*) the-envrun)))

;; na vrcholu *EXEC* je `envrun', provede se aplikace procedury
(define rule-eval-the-envrun
  (lambda ()
    (let* ((nargs (get-data (cadr *EXEC*)))  ; pocet argumentu
	   (env (caddr *EXEC*))              ; prostredi aplikace
	   (proc (list-ref *RSLT* nargs))    ; procedura
	   (args (list-head *RSLT* nargs)))  ; seznam argumentu

      ;; osetreni pro uzivatelske procedury, ostatni neovlivnuje
      (if (scm-user-procedure? proc)
	  (begin
	    (set! *EXEC*
		  `(,(procedure-body proc)
		    ,(make-env env
			       (make-bindings (procedure-arguments proc)
					      (reverse args)))
		    ,@(cdddr *EXEC*)))
	    (set! *RSLT* (list-tail *RSLT* (+ 1 nargs))))
	  (rule-eval-the-run)))))

;;;; VRACENI ELEMENTU Z RSLT ZPET NA EXEC

;; je na vrcholu *EXEC* element `return' a je *RSLT* neprazdny?
(define rule-eval-the-return?
  (lambda (*EXEC* *RSLT*)
    (and (eq? (car *EXEC*) the-return)
	 (not (null? *RSLT*)))))

;; vrat element z vrcholu *RSLT* na *EXEC*
(define rule-eval-the-return
  (lambda ()
    (set! *EXEC* (cons (car *RSLT*) (cdr *EXEC*)))
    (set! *RSLT* (cdr *RSLT*))))

;;;; PODMINENE VRACENI JEDNOHO Z ELEMENTU Z RSLT ZPET NA EXEC

;; je na vrcholu *EXEC* element `rotrue' a *RSLT* ma aspon 3 prvky?
(define rule-eval-the-rotrue?
  (lambda (*EXEC* *RSLT*)
    (and (eq? (car *EXEC*) the-rotrue)
	 (not (null? (cddr *RSLT*))))))

;; vrat element z vrcholu *RSLT* na *EXEC*
(define rule-eval-the-rotrue
  (lambda ()
    (set! *EXEC* (cons (if (eq? (car *RSLT*) scm-false)
			   (caddr *RSLT*)
			   (cadr *RSLT*))
		       (cdr *EXEC*)))
    (set! *RSLT* (cdddr *RSLT*))))

;;;; SKARTOVANI ELEMENTU Z *RSLT*

;; je na vrcholu *EXEC* element `discard'?
(define rule-eval-the-discard?
  (lambda (*EXEC* *RSLT*)
    (eq? (car *EXEC*) the-discard)))

;; skartuj jeden element z *RSLT*
(define rule-eval-the-discard
  (lambda ()
    (set! *EXEC* (cdr *EXEC*))
    (set! *RSLT* (cdr *RSLT*))))

;;;; DEFINICE VAZBY

;; je na vrcholu *EXEC* element `defasgn'?
(define rule-eval-the-defasgn?
  (lambda (*EXEC* *RSLT*)
    (eq? (car *EXEC*) the-defasgn)))

;; definuj novou vazbu v aktualnim prostredi
(define rule-eval-the-defasgn
  (lambda ()
    (let* ((value (car *RSLT*))
	   (symbol (cadr *RSLT*))
	   (env (cadr *EXEC*))
	   (result (lookup-env env symbol #f #f)))
      (if result
	  (pair-set-cdr! result value)
	  (pair-set-cdr!
	   (get-table env)
	   (make-pair (make-pair symbol value)
		      (pair-cdr (get-table env)))))
      (set! *EXEC* (cddr *EXEC*))
      (set! *RSLT* (cons the-undefined-value (cddr *RSLT*))))))

;;;; IMPERATIVNI ZMENA VAZBY

;; je na vrcholu *EXEC* element `setasgn'?
(define rule-eval-the-setasgn?
  (lambda (*EXEC* *RSLT*)
    (eq? (car *EXEC*) the-setasgn)))

;; imperativne zmen vazbu symbolu
(define rule-eval-the-setasgn
  (lambda ()
    (let* ((value (car *RSLT*))
	   (symbol (cadr *RSLT*))
	   (env (cadr *EXEC*)))
      (pair-set-cdr! (lookup-env env symbol #t #f) value)
      (set! *EXEC* (cddr *EXEC*))
      (set! *RSLT* (cons the-undefined-value (cddr *RSLT*))))))

;;;; VYHODNOCENI ELEMENTU NA SEBE SAMA

;; vyhodnoceni na sebe sama musi byt pouzito az na poslednim miste
;; podminka je vzdy pravidva
(define rule-eval-self?
  (lambda (*EXEC* *RSLT*)
    #t))

;; vyhdonoceni na sebe sama: z *EXEC* prehod element na *RSLT*
(define rule-eval-self
  (lambda ()
    (set! *RSLT* (cons (car *EXEC*) *RSLT*))
    (set! *EXEC* (cddr *EXEC*))))

;;;;
;;;;  VLASTNI ZASOBNIKOVY EVALUATOR
;;;;

;; tabulka pravidel pro evaluator
;; shoda stavu zasobniku s pravidly se testuje shora-dolu (zleva-doprava)
;; pocita se prvni shoda
(define scheme-evaluation-rules
  `((,rule-eval-symbol? . ,rule-eval-symbol)
    (,rule-eval-list? . ,rule-eval-list)
    (,rule-eval-the-inspect? . ,rule-eval-the-inspect)
    (,rule-eval-the-run? . ,rule-eval-the-run)
    (,rule-eval-the-return? . ,rule-eval-the-return)
    (,rule-eval-the-rotrue? . ,rule-eval-the-rotrue)
    (,rule-eval-the-discard? . ,rule-eval-the-discard)
    (,rule-eval-the-defasgn? . ,rule-eval-the-defasgn)
    (,rule-eval-the-setasgn? . ,rule-eval-the-setasgn)
    (,rule-eval-the-envrun? . ,rule-eval-the-envrun)
    (,rule-eval-self? . ,rule-eval-self)))

;; proved elementarni krok vyhodnoceni
(define scm-eval-step
  (lambda (table)
    (if (null? *EXEC*)
	#f
	(let match
	    ((table table))
	  (cond ((null? table) #f)
		(((caar table) *EXEC* *RSLT*)
		 (begin
		   ((cdar table))
		   #t))
		(else (match (cdr table))))))))

;; vyhodnot jeden vyraz a vrat vysledek jeho vyhodnoceni
(define scm-toplevel-eval
  (lambda (expr env)
    (set! *EXEC* (list expr env))
    (set! *RSLT* '())
    (let while
	((return #t))

      ;; vypis cinnost zasobniku
      (if return
	  (begin
	    (newline)
	    (print-stack *EXEC* "EXEC")
	    (print-stack *RSLT* "RSLT")))

      (if return
	  (while (scm-eval-step scheme-evaluation-rules))
	  (car *RSLT*)))))

;;;;
;;;;  POMOCNE PROCEDURY
;;;;

;; pro obecny typ volani apply sestavi seznam argumentu
(define apply-collect-arguments
  (lambda (args)
    (cond ((null? args) (error "; APPLY: argument missing"))
	  ((and (not (null? args)) (null? (cdr args)))
	   (scm-list->list (car args)))
	  (else (cons (car args) (apply-collect-arguments (cdr args)))))))

;;;;
;;;;  TOPLEVEL ENVIRONMENT
;;;;

;; vytvor prostredi, ktere je nejvys v hierarchii
(define scheme-toplevel-env
  (make-global-env
   `(
     ;; specialni forma IF
     (if . ,(make-manipulator
	     (make-specform
	      (lambda (e r)
		(list `(,(car r) ,(caddr e) ,the-rotrue ,@(cddr e))
		      (if (= (get-data (cadr e)) 2)
			  `(,(cadr r) ,the-undefined-value ,@(cdddr r))
			  `(,(cadr r) ,(caddr r) ,@(cddddr r))))))))

     ;; specialni forma LAMBDA
     (lambda . ,(make-specform
		 (lambda (env args . body)
		   (if (not (null? (cdr body)))
		       (make-procedure
			env args
			(make-pair (make-symbol 'begin)
				   (let iter ((ar body))
				     (if (null? (cdr ar))
					 (make-pair (car ar) the-empty-list)
					 (make-pair (car ar)
						    (iter (cdr ar)))))))
		       (make-procedure env args (car body))))))

     ;; specialni forma QUOTE
     (quote . ,(make-specform (lambda (env elem) elem)))

     ;; specialni forma THE-ENVIRONMENT
     (the-environment . ,(make-specform (lambda (env) env)))

     ;; specialni forma BEGIN
     (begin . ,(make-manipulator
		(make-specform
		 (lambda (e r)
		   (list
		    `(,@(apply append
			       (let iter
				   ((body (list-head r (get-data (cadr e)))))
				 (cond ((null? body)
					`((,the-undefined-value ,(caddr e))))
				       ((null? (cdr body))
					`((,(car body) ,(caddr e))))
				       (else (cons `(,(car body)
						     ,(caddr e)
						     ,the-discard)
						   (iter (cdr body)))))))
		      ,@(cdddr e))
		    `(,@(list-tail r (+ 1 (get-data (cadr e))))))))))

     ;; specialni forma DEFINE
     (define . ,(make-manipulator
		 (make-specform
		  (lambda (e r)
		    (list `(,(cadr r) ,(caddr e) ,the-defasgn ,@(cddr e))
			  `(,(car r) ,@(cdddr r)))))))

     ;; specialni forma SET!
     (set! . ,(make-manipulator
	       (make-specform
		(lambda (e r)
		  (list `(,(cadr r) ,(caddr e) ,the-setasgn ,@(cddr e))
			`(,(car r) ,@(cdddr r)))))))
     
     ;; konstanty
     (pi . ,(make-number (* 4 (atan 1))))

     ;; procedury pro praci s cisly (zakomentovane procedury jsou v R5RS, ale
     ;; nejsou implementovany ve vsech dostupnych interpretech Scheme)
     (* . ,(wrap-primitive *))
     (+ . ,(wrap-primitive +))
     (- . ,(wrap-primitive -))
     (/ . ,(wrap-primitive /))
     (< . ,(wrap-primitive <))
     (<= . ,(wrap-primitive <=))
     (= . ,(wrap-primitive =))
     (> . ,(wrap-primitive >))
     (>= . ,(wrap-primitive >=))
     (abs . ,(wrap-primitive abs))
     (acos . ,(wrap-primitive acos))
;     (angle . ,(wrap-primitive angle))
     (asin . ,(wrap-primitive asin))
     (atan . ,(wrap-primitive atan))
     (ceiling . ,(wrap-primitive ceiling))
     (cos . ,(wrap-primitive cos))
;     (denominator . ,(wrap-primitive denominator))
     (even? . ,(wrap-primitive even?))
     (exact->inexact . ,(wrap-primitive exact->inexact))
     (exact? . ,(wrap-primitive exact?))
     (exp . ,(wrap-primitive exp))
     (expt . ,(wrap-primitive expt))
     (floor . ,(wrap-primitive floor))
     (gcd . ,(wrap-primitive gcd))
;     (imag-part . ,(wrap-primitive imag-part))
     (inexact->exact . ,(wrap-primitive inexact->exact))
     (inexact? . ,(wrap-primitive inexact?))
     (lcm . ,(wrap-primitive lcm))
     (log . ,(wrap-primitive log))
;     (magnitude . ,(wrap-primitive magnitude))
;     (make-polar . ,(wrap-primitive make-polar))
;     (make-rectangular . ,(wrap-primitive make-rectangular))
     (max . ,(wrap-primitive max))
     (min . ,(wrap-primitive min))
     (modulo . ,(wrap-primitive modulo))
     (negative? . ,(wrap-primitive negative?))
;     (numerator . ,(wrap-primitive numerator))
     (odd? . ,(wrap-primitive odd?))
     (positive? . ,(wrap-primitive positive?))
     (quotient . ,(wrap-primitive quotient))
;     (rationalize . ,(wrap-primitive rationalize))
;     (real-part . ,(wrap-primitive real-part))
     (remainder . ,(wrap-primitive remainder))
     (round . ,(wrap-primitive round))
     (sin . ,(wrap-primitive sin))
     (sqrt . ,(wrap-primitive sqrt))
     (tan . ,(wrap-primitive tan))
     (truncate . ,(wrap-primitive truncate))
     (zero? . ,(wrap-primitive zero?))

     ;; prace s pary
     (cons . ,(make-primitive make-pair))
     (car . ,(make-primitive pair-car))
     (cdr . ,(make-primitive pair-cdr))
     (set-car! . ,(make-primitive pair-set-car!))
     (set-cdr! . ,(make-primitive pair-set-cdr!))

     ;; negace
     (not . ,(make-primitive
	      (lambda (elem)
		(if (equal? elem scm-false)
		    scm-true
		    scm-false))))

     ;; predikaty
     (boolean? . ,(wrap-predicate scm-boolean?))
     (form? . ,(wrap-predicate scm-form?))
     (null? . ,(wrap-predicate scm-null?))
     (number? . ,(wrap-predicate scm-number?))
     (pair? . ,(wrap-predicate scm-pair?))
     (primitive? . ,(wrap-predicate scm-primitive?))
     (procedure? . ,(wrap-predicate scm-procedure?))
     (specform? . ,(wrap-predicate scm-specform?))
     (symbol? . ,(wrap-predicate scm-symbol?))
     (undefined? . ,(wrap-predicate scm-undefined?))
     (user-procedure? . ,(wrap-predicate scm-user-procedure?))
     (void? . ,(wrap-predicate scm-void?))
     (eq? . ,(wrap-predicate scm-eq?))
     
     ;; sektoru uzivatelskych procedur
     (procedure-environment . ,(make-primitive procedure-environment))
     (procedure-arguments . ,(make-primitive procedure-arguments))
     (procedure-body . ,(make-primitive procedure-body))

     ;; procedura environment-parent (vrat lexikalniho predka daneho prostredi)
     (environment-parent . ,(make-primitive get-pred))

     ;; procedura environment->list (vrat tabulku s vazbami v danem prosredi)
     (environment->list . ,(make-primitive
			    (lambda (elem)
			      (if (equal? elem scm-false)
				  scm-false
				  (get-table elem)))))

     ;; manipulator eval
     (eval . ,(make-manipulator
	       (make-primitive
		(lambda (e r)
		  (list `(,(cadr r) ,(car r) ,@(cdddr e))
			(cdddr r))))))

     ;; manipulator apply
     (apply . ,(make-manipulator
		(make-primitive
		 (lambda (e r)
		   (let* ((args
			   (reverse
			    (apply-collect-arguments
			     (reverse
			      (list-head r (- (get-data (cadr e)) 1))))))
			  (nargs (length args)))
		     (list `(,the-run ,(make-number nargs) ,@(cddr e))
			   `(,@args
			     ,(list-ref r (- (get-data (cadr e)) 1))
			     ,@(list-tail r (+ (get-data (cadr e)) 1)))))))))

     ;; procedura macro + vracec transformatoru
     (macro . ,(make-primitive make-macro))
     (macro-transformer . ,(make-primitive get-data))

     ;; generovane symboly
     (gensym . ,(make-primitive make-generated-symbol))

     ;; manipulator apply, kteremu lze predat prostredi predka
     (env-apply . ,(make-manipulator
		(make-primitive
		 (lambda (e r)
		   (let* ((args
			   (reverse
			    (apply-collect-arguments
			     (reverse
			      (list-head r (- (get-data (cadr e)) 2))))))
			  (nargs (length args)))
		     (list `(,the-envrun ,(make-number nargs) ,@(cddr e))
			   `(,@args
			     ,(list-ref r (- (get-data (cadr e)) 1))
			     ,@(list-tail r (+ (get-data (cadr e)) 1)))))))))

     ;; manipulator call/cc
     (call/cc . ,(make-manipulator
		  (make-primitive
		   (lambda (e r)
		     (let ((continuation
			    (let ((saved-EXEC (cdddr *EXEC*))
				  (saved-RSLT (cddr *RSLT*)))
			      (make-manipulator
			       (make-primitive
				(lambda (e r)
				  (list saved-EXEC
					(cons (if (= (get-data (cadr e)) 0)
						  the-undefined-value
						  (car r))
					      saved-RSLT)))))))
			   (receiver (car r)))
		       (list `(,the-run ,(make-number 1) ,@(cddr e))
			     `(,continuation
			       ,receiver
			       ,@(cddr r))))))))
     
     ;; procedura lookup-env
     (lookup-env . ,(make-primitive
		     (lambda (env symbol . nonlocal)
		       (lookup-env env symbol
				   (or (null? nonlocal)
				       (equal? (car nonlocal) scm-true))
				   scm-false))))

     )))

;; procedura load: vyhodnoceni vyrazu v souboru (slouzi k natazeni knihovny)
(define scm-load
  (lambda (file-name env)
    (let ((port (open-input-file file-name)))
      (let next ()
	(let ((expr (read port)))
	  (if (not (eof-object? expr))
	      (begin
		(scm-toplevel-eval (expr->intern expr) env)
		(next)))))
      (close-input-port port))))

;; vrat nazev typu jako retezec s prvnim pismenem velkym
(define type-name
  (lambda (elem)
    (let ((strl (string->list (symbol->string (get-type-tag elem)))))
      (list->string (cons (char-upcase (car strl)) (cdr strl))))))

;;;;
;;;;  NOVY REPL
;;;;

;; cyklus REPL
(define scm-repl
  (lambda ()
    (let ((glob-env (make-env scheme-toplevel-env the-empty-list)))
      (scm-load "includes.scm" glob-env)
      (let loop ()
	(display "]=> ")
	(let ((elem (scm-read)))
	  (if (not elem)
	      'bye-bye
	      (let ((result (scm-toplevel-eval elem glob-env)))
		(newline)
		(display (string-append "; " (type-name result)	": "))
		(scm-print result)
		(newline)
		(newline)
		(loop))))))))

;; spust REPL
(scm-repl)
