
;;  SCHEME IN SCHEME (STACK-BASED IMPLEMENTATION)
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

(define caar (lambda (x) (car (car x))))
(define cadr (lambda (x) (car (cdr x))))
(define cdar (lambda (x) (cdr (car x))))
(define cddr (lambda (x) (cdr (cdr x))))
(define caaar (lambda (x) (car (car (car x)))))
(define caadr (lambda (x) (car (car (cdr x)))))
(define cadar (lambda (x) (car (cdr (car x)))))
(define caddr (lambda (x) (car (cdr (cdr x)))))
(define cdaar (lambda (x) (cdr (car (car x)))))
(define cdadr (lambda (x) (cdr (car (cdr x)))))
(define cddar (lambda (x) (cdr (cdr (car x)))))
(define cdddr (lambda (x) (cdr (cdr (cdr x)))))
(define caaaar (lambda (x) (car (car (car (car x))))))
(define caaadr (lambda (x) (car (car (car (cdr x))))))
(define caadar (lambda (x) (car (car (cdr (car x))))))
(define caaddr (lambda (x) (car (car (cdr (cdr x))))))
(define cadaar (lambda (x) (car (cdr (car (car x))))))
(define cadadr (lambda (x) (car (cdr (car (cdr x))))))
(define caddar (lambda (x) (car (cdr (cdr (car x))))))
(define cadddr (lambda (x) (car (cdr (cdr (cdr x))))))
(define cdaaar (lambda (x) (cdr (car (car (car x))))))
(define cdaadr (lambda (x) (cdr (car (car (cdr x))))))
(define cdadar (lambda (x) (cdr (car (cdr (car x))))))
(define cdaddr (lambda (x) (cdr (car (cdr (cdr x))))))
(define cddaar (lambda (x) (cdr (cdr (car (car x))))))
(define cddadr (lambda (x) (cdr (cdr (car (cdr x))))))
(define cdddar (lambda (x) (cdr (cdr (cdr (car x))))))
(define cddddr (lambda (x) (cdr (cdr (cdr (cdr x))))))

(define length
  (lambda (l)
    (if (null? l)
        0
        (+ 1 (length (cdr l))))))

(define list
  (lambda list list))

(define list?
  (lambda (l)
    (if (null? l)
        #t
        (if (pair? l)
	    (list? (cdr l))
	    #f))))

(define list-tail
  (lambda (l skip)
    (if (= skip 0)
        l
        (list-tail (cdr l) (- skip 1)))))

(define list-ref
  (lambda (l index)
    (car (list-tail l index))))

(define build-list
  (lambda (n f)
    (begin
      (define iter
	(lambda (i)
	  (if (= i n)
	      '()
	      (cons (f i) (iter (+ i 1))))))
      (iter 0))))

(define append
  (lambda lists
    (if (null? lists)
	'()
	(if (null? (car lists))
	    (apply append (cdr lists))
	    (cons (car (car lists))
		  (apply append
			 (cons (cdr (car lists))
			       (cdr lists))))))))

(define reverse
  ((lambda ()
     (begin
       (define iter
	 (lambda (l accum)
	   (if (null? l)
	       accum
	       (iter (cdr l) (cons (car l) accum)))))
       (lambda (l)
	 (iter l '()))))))

(define map
  ((lambda ()
     (begin
       (define map1
	 (lambda (f l)
	   (if (null? l)
	       '()
	       (cons (f (car l))
		     (map1 f (cdr l))))))
       (lambda (f . lists)
	 (if (null? (car lists))
	     '()
	     (cons (apply f (map1 car lists))
		   (apply map f (map1 cdr lists)))))))))

(define for-each
  (lambda (f . lists)
    (if (not (null? (car lists)))
	(begin
	  (apply f (map car lists))
	  (apply for-each f (map cdr lists))))))

(define foldr
  (lambda (f basis . lists)
    (if (null? (car lists))
        basis
        (apply f
	       (append (map car lists)
		       (list (apply foldr f basis (map cdr lists))))))))

(define foldl
  ((lambda ()
     (begin
       (define iter
	 (lambda (f lists accum)
	   (if (null? (car lists))
	       accum
	       (iter f
		     (map cdr lists)
		     (apply f (append (map car lists)
				      (list accum)))))))
       (lambda (f basis . lists)
	 (iter f lists basis))))))

(define eqv?
  (lambda (x y)
    (if (and (number? x)
	     (number? y)
	     (or (and (exact? x)
		      (exact? y))
		 (and (inexact? x)
		      (inexact? y))))
        (= x y)
        (eq? x y))))

(define equal?
  (lambda (x y)
    (or (eqv? x y)
        (and (pair? x)
	     (pair? y)
             (equal? (car x) (car y))
             (equal? (cdr x) (cdr y))))))

(define make-finder
  (lambda (equality-predicate? return)
    (begin
      (define finder-proc
	(lambda (elem list)
	  (if (null? list)
	      #f
	      (if (equality-predicate? elem (car list))
		  (return list)
		  (finder-proc elem (cdr list))))))
      finder-proc)))

(define memq (make-finder eq? (lambda (x) x)))
(define memv (make-finder eqv? (lambda (x) x)))
(define member (make-finder equal? (lambda (x) x)))
(define assq (make-finder (lambda (x y) (eq? x (car y))) car))
(define assv (make-finder (lambda (x y) (eqv? x (car y)))  car))
(define assoc (make-finder (lambda (x y) (equal? x (car y))) car))

(define quasiquote
  (macro
    (lambda (expr)
      (if (not (list? expr))
	  (list 'quote expr)
	  (apply
	   list 'apply 'append
	   (append
	    (map (lambda (expr)
		   (if (list? expr)
		       (if (null? expr)
			   '(list (quote ()))
			   (if (eq? (car expr) 'unquote)
			       (list 'list (cadr expr))
			       (if (eq? (car expr) 'unquote-splicing)
				   (cadr expr)
				   (if (eq? (car expr) 'quasiquote)
				       (list 'list (list 'quote expr))
				       (list 'list (list 'quasiquote expr))))))
		       (list 'list (list 'quote expr))))
		 expr)
	    '((quote ()))))))))

(define let
  (macro
    (lambda args
      (if (symbol? (car args))
	  `((lambda ()
	      (begin
		(define ,(car args)
		  (lambda ,(map car (cadr args))
		    (begin ,@(cddr args)))))
	      (,(car args) ,@(map cadr (cadr args)))))
	  `((lambda ,(map car (car args))
	      (begin ,@(cdr args)))
	    ,@(map cadr (car args)))))))

(define let*
  (macro
    (lambda (assgn . body)
      (if (null? assgn)
	  `((lambda () ,@body))
	  `((lambda (,(caar assgn))
	      (let* ,(cdr assgn)
		,@body))
	    ,(cadar assgn))))))

(define letrec
  (macro
    (lambda (assgn . body)
      `((lambda ()
	  ,@(map (lambda (i)
		   `(define ,(car i) ,(cadr i)))
		 assgn)
	  ,@body)))))

(define and
  (macro
    (lambda args
      (if (null? args)
	  #t
	  (if (null? (cdr args))
	      (car args)
	      `(if ,(car args) 
		   (and ,@(cdr args))
		   #f))))))

(define or
  (macro
    (lambda args
      (if (null? args)
	  #f
	  (if (null? (cdr args))
	      (car args)
	      (let ((result (gensym)))
		`(let ((,result ,(car args)))
		   (if ,result
		       ,result
		       (or ,@(cdr args))))))))))

(define cond
  (macro
    (lambda clist
      (let ((symbol (gensym)))
	(if (null? clist)
	    `(if #f #f)
	    (if (equal? (caar clist) 'else)
		`(begin ,@(cdar clist))
		`(let ((,symbol ,(caar clist)))
		   (if ,symbol
		       ,(if (null? (cdar clist))
			    symbol
			    (if (equal? (cadar clist) '=>)
				`(,(caddar clist) ,symbol)
				`(begin ,@(cdar clist))))
		       (cond ,@(cdr clist))))))))))

(define case
  (macro
    (lambda (value . clist)
      (let ((result (gensym)))
	`(let ((,result ,value))
	   (cond ,@(map (lambda (x)
			  (if (list? (car x))
			      `((member ,result ',(car x)) ,(cadr x))
			      `(else ,(cadr x))))
			clist)))))))

(define do
  (macro
    (lambda (binding condition . body)
      (let ((loop-name (gensym)))
	`(let ,loop-name
	   ,(map (lambda (x)
		   (list (car x) (cadr x)))
		 binding)
	   (if ,(car condition)
	       (begin ,@(cdr condition))
	       (begin ,@body
		      (,loop-name ,@(map caddr binding)))))))))

(define the-environment
  (macro
    (lambda ()
      '(procedure-environment (lambda () #f)))))

(define dyn-apply
  (macro
    (lambda (proc . args)
      `(env-apply ,proc (the-environment) ,@args))))
